==> pylogram-0.5.1-py3-none-any.whl: pylogram/utils.py <==
#  Pylogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-2023 Dan <https://github.com/delivrance>
#  Copyright (C) 2023-2024 Pylakey <https://github.com/pylakey>
#
#  This file is part of Pylogram.
#
#  Pylogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pylogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pylogram.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import base64
import contextlib
import hashlib
import logging
import os
import re
import struct
import sys
import time
import typing
from datetime import datetime
from datetime import timezone
from getpass import getpass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pylogram
from pylogram import enums
from pylogram import raw
from pylogram import types
from pylogram.file_id import DOCUMENT_TYPES
from pylogram.file_id import FileId
from pylogram.file_id import FileType
from pylogram.file_id import PHOTO_TYPES


def _read_input(prompt: str = "") -> str:
    print(prompt, end=" ", flush=True)
    return sys.stdin.readline()


async def ainput(prompt: str = "", *, hide: bool = False) -> str:
    if len(prompt):
        prompt = prompt.strip()

    if hide:
        stdin = await asyncio.to_thread(getpass, prompt=prompt, stream=None)
    else:
        stdin = await asyncio.to_thread(_read_input, prompt)

    return stdin.strip()


def get_input_media_from_file_id(
        file_id: str,
        expected_file_type: FileType = None,
        ttl_seconds: int = None
) -> Union["raw.types.InputMediaPhoto", "raw.types.InputMediaDocument"]:
    try:
        decoded = FileId.decode(file_id)
    except Exception:
        raise ValueError(
            f'Failed to decode "{file_id}". The value does not represent an existing local file, '
            f"HTTP URL, or valid file id."
        )

    file_type = decoded.file_type

    if expected_file_type is not None and file_type != expected_file_type:
        raise ValueError(f"Expected {expected_file_type.name}, got {file_type.name} file id instead")

    if file_type in (FileType.THUMBNAIL, FileType.CHAT_PHOTO):
        raise ValueError(f"This file id can only be used for download: {file_id}")

    if file_type in PHOTO_TYPES:
        return raw.types.InputMediaPhoto(
            id=raw.types.InputPhoto(
                id=decoded.media_id,
                access_hash=decoded.access_hash,
                file_reference=decoded.file_reference
            ),
            ttl_seconds=ttl_seconds
        )

    if file_type in DOCUMENT_TYPES:
        return raw.types.InputMediaDocument(
            id=raw.types.InputDocument(
                id=decoded.media_id,
                access_hash=decoded.access_hash,
                file_reference=decoded.file_reference
            ),
            ttl_seconds=ttl_seconds
        )

    raise ValueError(f"Unknown file id: {file_id}")


async def parse_messages(
        client,
        messages: "raw.types.messages.Messages",
        replies: int = 1
) -> List["types.Message"]:
    users = {i.id: i for i in messages.users}
    chats = {i.id: i for i in messages.chats}

    if not messages.messages:
        return types.List()

    parsed_messages = []

    for message in messages.messages:
        parsed_messages.append(await types.Message._parse(client, message, users, chats, replies=0))

    if replies:
        messages_with_replies = {
            i.id: i.reply_to.reply_to_msg_id
            for i in messages.messages
            if not isinstance(i, raw.types.MessageEmpty) and isinstance(i.reply_to, raw.types.MessageReplyHeader)
        }

        if messages_with_replies:
            # We need a chat id, but some messages might be empty (no chat attribute available)
            # Scan until we find a message with a chat available (there must be one, because we are fetching replies)
            for m in parsed_messages:
                if m.chat:
                    chat_id = m.chat.id
                    break
            else:
                chat_id = 0

            reply_messages = await client.get_messages(
                chat_id,
                reply_to_message_ids=messages_with_replies.keys(),
                replies=replies - 1
            )

            for message in parsed_messages:
                reply_id = messages_with_replies.get(message.id, None)

                for reply in reply_messages:
                    if reply.id == reply_id:
                        message.reply_to_message = reply

    return types.List(parsed_messages)


def parse_deleted_messages(client, update) -> List["types.Message"]:
    messages = update.messages
    channel_id = getattr(update, "channel_id", None)

    parsed_messages = []

    for message in messages:
        parsed_messages.append(
            types.Message(
                id=message,
                chat=types.Chat(
                    id=get_channel_id(channel_id),
                    type=enums.ChatType.CHANNEL,
                    client=client
                ) if channel_id is not None else None,
                client=client,
                raw_message=message
            )
        )

    return types.List(parsed_messages)


def pack_inline_message_id(msg_id: "raw.base.InputBotInlineMessageID"):
    if isinstance(msg_id, raw.types.InputBotInlineMessageID):
        inline_message_id_packed = struct.pack(
            "<iqq",
            msg_id.dc_id,
            msg_id.id,
            msg_id.access_hash
        )
    else:
        inline_message_id_packed = struct.pack(
            "<iqiq",
            msg_id.dc_id,
            msg_id.owner_id,
            msg_id.id,
            msg_id.access_hash
        )

    return base64.urlsafe_b64encode(inline_message_id_packed).decode().rstrip("=")


def unpack_inline_message_id(inline_message_id: str) -> "raw.base.InputBotInlineMessageID":
    padded = inline_message_id + "=" * (-len(inline_message_id) % 4)
    decoded = base64.urlsafe_b64decode(padded)

    if len(decoded) == 20:
        unpacked = struct.unpack("<iqq", decoded)

        return raw.types.InputBotInlineMessageID(
            dc_id=unpacked[0],
            id=unpacked[1],
            access_hash=unpacked[2]
        )
    else:
        unpacked = struct.unpack("<iqiq", decoded)

        return raw.types.InputBotInlineMessageID64(
            dc_id=unpacked[0],
            owner_id=unpacked[1],
            id=unpacked[2],
            access_hash=unpacked[3]
        )


MIN_CHANNEL_ID = -1002147483647
MAX_CHANNEL_ID = -1000000000000
MIN_CHAT_ID = -2147483647
MAX_USER_ID_OLD = 2147483647
MAX_USER_ID = 999999999999


def get_raw_peer_id(peer: raw.base.Peer) -> Optional[int]:
    """Get the raw peer id from a Peer object"""
    if isinstance(peer, raw.types.PeerUser):
        return peer.user_id

    if isinstance(peer, raw.types.PeerChat):
        return peer.chat_id

    if isinstance(peer, raw.types.PeerChannel):
        return peer.channel_id

    return None


def get_peer_id(peer: raw.base.Peer | raw.base.InputPeer) -> int:
    """Get the non-raw peer id from a Peer object"""
    if isinstance(peer, (raw.types.PeerUser, raw.types.InputPeerUser)):
        return peer.user_id

    if isinstance(peer, (raw.types.PeerChat, raw.types.InputPeerChat)):
        return -peer.chat_id

    if isinstance(peer, (raw.types.PeerChannel, raw.types.InputPeerChannel)):
        return MAX_CHANNEL_ID - peer.channel_id

    if isinstance(peer, raw.types.InputPeerChannel):
        return MAX_CHANNEL_ID - peer.channel_id

    raise ValueError(f"Peer type invalid: {peer}")


def get_peer_type(peer_id: int) -> str:
    if peer_id < 0:
        if MIN_CHAT_ID <= peer_id:
            return "chat"

        if MIN_CHANNEL_ID <= peer_id < MAX_CHANNEL_ID:
            return "channel"
    elif 0 < peer_id <= MAX_USER_ID:
        return "user"

    raise ValueError(f"Peer id invalid: {peer_id}")


def get_channel_id(peer_id: int) -> int:
    return MAX_CHANNEL_ID - peer_id


def btoi(b: bytes) -> int:
    return int.from_bytes(b, "big")


def itob(i: int) -> bytes:
    return i.to_bytes(256, "big")


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def xor(a: bytes, b: bytes) -> bytes:
    return bytes(i ^ j for i, j in zip(a, b))


def compute_password_hash(
        algo: raw.types.PasswordKdfAlgoSHA256SHA256PBKDF2HMACSHA512iter100000SHA256ModPow,
        password: str
) -> bytes:
    hash1 = sha256(algo.salt1 + password.encode() + algo.salt1)
    hash2 = sha256(algo.salt2 + hash1 + algo.salt2)
    hash3 = hashlib.pbkdf2_hmac("sha512", hash2, algo.salt1, 100000)

    return sha256(algo.salt2 + hash3 + algo.salt2)


# noinspection PyPep8Naming
def compute_password_check(
        r: raw.types.account.Password,
        password: str
) -> raw.types.InputCheckPasswordSRP:
    algo = r.current_algo

    p_bytes = algo.p
    p = btoi(algo.p)

    g_bytes = itob(algo.g)
    g = algo.g

    B_bytes = r.srp_B
    B = btoi(B_bytes)

    srp_id = r.srp_id

    x_bytes = compute_password_hash(algo, password)
    x = btoi(x_bytes)

    g_x = pow(g, x, p)

    k_bytes = sha256(p_bytes + g_bytes)
    k = btoi(k_bytes)

    kg_x = (k * g_x) % p

    while True:
        a_bytes = os.urandom(256)
        a = btoi(a_bytes)

        A = pow(g, a, p)
        A_bytes = itob(A)

        u = btoi(sha256(A_bytes + B_bytes))

        if u > 0:
            break

    g_b = (B - kg_x) % p

    ux = u * x
    a_ux = a + ux
    S = pow(g_b, a_ux, p)
    S_bytes = itob(S)

    K_bytes = sha256(S_bytes)

    M1_bytes = sha256(
        xor(sha256(p_bytes), sha256(g_bytes))
        + sha256(algo.salt1)
        + sha256(algo.salt2)
        + A_bytes
        + B_bytes
        + K_bytes
    )

    return raw.types.InputCheckPasswordSRP(srp_id=srp_id, A=A_bytes, M1=M1_bytes)


async def parse_text_entities(
        client: "pylogram.Client",
        text: str,
        parse_mode: enums.ParseMode,
        entities: List["types.MessageEntity"]
) -> Dict[str, Union[str, List[raw.base.MessageEntity]]]:
    if entities:
        # Inject the client instance because parsing user mentions requires it
        for entity in entities:
            entity._client = client

        text, entities = text, [await entity.write() for entity in entities] or None
    else:
        text, entities = (await client.parser.parse(text, parse_mode)).values()

    return {
        "message": text,
        "entities": entities
    }


def zero_datetime() -> datetime:
    return datetime.fromtimestamp(0, timezone.utc)


def timestamp_to_datetime(ts: Optional[int]) -> Optional[datetime]:
    return datetime.fromtimestamp(ts) if ts else None


def datetime_to_timestamp(dt: Optional[datetime]) -> Optional[int]:
    return int(dt.timestamp()) if dt else None


@contextlib.asynccontextmanager
async def async_perf_counter(context: str = ""):
    start = time.perf_counter()

    try:
        yield
    finally:
        end = time.perf_counter()
        elapsed_time = end - start
        logging.info(f"[{context}] Elapsed time: {elapsed_time}")


def parse_username(username_string: str) -> str:
    username_string = username_string.lower()
    # Remove all valid user url components
    # like (https://username.t.me, https://t.me/username, https://telegram.me/username)
    username_string = re.sub(r"(http(s)?://)|(\.?t\.me/?)", "", username_string)
    # Remove all non-username characters
    username_string = re.sub(r"[^a-z0-9_]", "", username_string)

    return username_string


def parse_phone_number(phone_number: int | str) -> str | None:
    phone_number = re.sub(r"\D", "", str(phone_number))

    # Validate phone number format according to ISO 3166-1 alpha-2
    if not re.match(r"^\d{5,15}$", phone_number):
        return None

    return phone_number


def calculate_pagination_hash(numbers: List[int]) -> int:
    acc = 0

    for number in numbers:
        acc ^= acc >> 21
        acc ^= acc << 35
        acc ^= acc >> 4
        acc = (acc + number) & 0xffffffffffffffff  # To let it not overflow a 64-bit integer

    return acc


def is_tl_object_of_base_type(value: pylogram.raw.core.TLObject, base_type: Any | tuple[Any]) -> bool:
    if isinstance(base_type, tuple):
        return any(is_tl_object_of_base_type(value, t) for t in base_type)

    if base_type is None or base_type is Any:
        return True

    try:
        if isinstance(value, base_type):
            return True
    except TypeError:
        pass

    try:
        if issubclass(type(value), base_type):
            return True
    except TypeError:
        pass

    if base_type is type(value):
        return True

    origin = typing.get_origin(base_type)

    if origin is Union:
        for arg in typing.get_args(base_type):
            if is_tl_object_of_base_type(value, arg):
                return True

        return False

    return False


CHAT_LIST_INVITE_LINK_PREFIX: str = "https://t.me/addlist/"


def slug_to_chat_list_invite_link(slug: str) -> str:
    return f"{self.CHAT_LIST_INVITE_LINK_PREFIX}{slug}"


def chat_list_invite_link_to_slug(link: str) -> str:
    return link.replace(self.CHAT_LIST_INVITE_LINK_PREFIX, '')

==> pylogram-0.6.3-py3-none-any.whl: pylogram/utils.py <==
#  Pylogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-2023 Dan <https://github.com/delivrance>
#  Copyright (C) 2023-2024 Pylakey <https://github.com/pylakey>
#
#  This file is part of Pylogram.
#
#  Pylogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pylogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pylogram.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import base64
import contextlib
import hashlib
import logging
import os
import re
import struct
import sys
import time
import typing
from datetime import datetime
from datetime import timezone
from getpass import getpass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pylogram
from pylogram import enums
from pylogram import raw
from pylogram import types
from pylogram.file_id import DOCUMENT_TYPES
from pylogram.file_id import FileId
from pylogram.file_id import FileType
from pylogram.file_id import PHOTO_TYPES


def _read_input(prompt: str = "") -> str:
    print(prompt, end=" ", flush=True)
    return sys.stdin.readline()


async def ainput(prompt: str = "", *, hide: bool = False) -> str:
    if len(prompt):
        prompt = prompt.strip()

    if hide:
        stdin = await asyncio.to_thread(getpass, prompt=prompt, stream=None)
    else:
        stdin = await asyncio.to_thread(_read_input, prompt)

    return stdin.strip()


def get_input_media_from_file_id(
        file_id: str,
        expected_file_type: FileType = None,
        ttl_seconds: int = None
) -> Union["raw.types.InputMediaPhoto", "raw.types.InputMediaDocument"]:
    try:
        decoded = FileId.decode(file_id)
    except Exception:
        raise ValueError(
            f'Failed to decode "{file_id}". The value does not represent an existing local file, '
            f"HTTP URL, or valid file id."
        )

    file_type = decoded.file_type

    if expected_file_type is not None and file_type != expected_file_type:
        raise ValueError(f"Expected {expected_file_type.name}, got {file_type.name} file id instead")

    if file_type in (FileType.THUMBNAIL, FileType.CHAT_PHOTO):
        raise ValueError(f"This file id can only be used for download: {file_id}")

    if file_type in PHOTO_TYPES:
        return raw.types.InputMediaPhoto(
            id=raw.types.InputPhoto(
                id=decoded.media_id,
                access_hash=decoded.access_hash,
                file_reference=decoded.file_reference
            ),
            ttl_seconds=ttl_seconds
        )

    if file_type in DOCUMENT_TYPES:
        return raw.types.InputMediaDocument(
            id=raw.types.InputDocument(
                id=decoded.media_id,
                access_hash=decoded.access_hash,
                file_reference=decoded.file_reference
            ),
            ttl_seconds=ttl_seconds
        )

    raise ValueError(f"Unknown file id: {file_id}")


async def parse_messages(
        client,
        messages: "raw.types.messages.Messages",
        replies: int = 1
) -> List["types.Message"]:
    users = {i.id: i for i in messages.users}
    chats = {i.id: i for i in messages.chats}

    if not messages.messages:
        return types.List()

    parsed_messages = []

    for message in messages.messages:
        parsed_messages.append(await types.Message._parse(client, message, users, chats, replies=0))

    if replies:
        messages_with_replies = {
            i.id: i.reply_to.reply_to_msg_id
            for i in messages.messages
            if not isinstance(i, raw.types.MessageEmpty) and isinstance(i.reply_to, raw.types.MessageReplyHeader)
        }

        if messages_with_replies:
            # We need a chat id, but some messages might be empty (no chat attribute available)
            # Scan until we find a message with a chat available (there must be one, because we are fetching replies)
            for m in parsed_messages:
                if m.chat:
                    chat_id = m.chat.id
                    break
            else:
                chat_id = 0

            reply_messages = await client.get_messages(
                chat_id,
                reply_to_message_ids=messages_with_replies.keys(),
                replies=replies - 1
            )

            for message in parsed_messages:
                reply_id = messages_with_replies.get(message.id, None)

                for reply in reply_messages:
                    if reply.id == reply_id:
                        message.reply_to_message = reply

    return types.List(parsed_messages)


def parse_deleted_messages(client, update) -> List["types.Message"]:
    messages = update.messages
    channel_id = getattr(update, "channel_id", None)

    parsed_messages = []

    for message in messages:
        parsed_messages.append(
            types.Message(
                id=message,
                chat=types.Chat(
                    id=get_channel_id(channel_id),
                    type=enums.ChatType.CHANNEL,
                    client=client
                ) if channel_id is not None else None,
                client=client,
                raw_message=message
            )
        )

    return types.List(parsed_messages)


def pack_inline_message_id(msg_id: "raw.base.InputBotInlineMessageID"):
    if isinstance(msg_id, raw.types.InputBotInlineMessageID):
        inline_message_id_packed = struct.pack(
            "<iqq",
            msg_id.dc_id,
            msg_id.id,
            msg_id.access_hash
        )
    else:
        inline_message_id_packed = struct.pack(
            "<iqiq",
            msg_id.dc_id,
            msg_id.owner_id,
            msg_id.id,
            msg_id.access_hash
        )

    return base64.urlsafe_b64encode(inline_message_id_packed).decode().rstrip("=")


def unpack_inline_message_id(inline_message_id: str) -> "raw.base.InputBotInlineMessageID":
    padded = inline_message_id + "=" * (-len(inline_message_id) % 4)
    decoded = base64.urlsafe_b64decode(padded)

    if len(decoded) == 20:
        unpacked = struct.unpack("<iqq", decoded)

        return raw.types.InputBotInlineMessageID(
            dc_id=unpacked[0],
            id=unpacked[1],
            access_hash=unpacked[2]
        )
    else:
        unpacked = struct.unpack("<iqiq", decoded)

        return raw.types.InputBotInlineMessageID64(
            dc_id=unpacked[0],
            owner_id=unpacked[1],
            id=unpacked[2],
            access_hash=unpacked[3]
        )


MIN_CHANNEL_ID = -1002147483647
MAX_CHANNEL_ID = -1000000000000
MIN_CHAT_ID = -2147483647
MAX_USER_ID_OLD = 2147483647
MAX_USER_ID = 999999999999


def get_raw_peer_id(peer: raw.base.Peer) -> Optional[int]:
    """Get the raw peer id from a Peer object"""
    if isinstance(peer, raw.types.PeerUser):
        return peer.user_id

    if isinstance(peer, raw.types.PeerChat):
        return peer.chat_id

    if isinstance(peer, raw.types.PeerChannel):
        return peer.channel_id

    return None


def get_peer_id(peer: raw.base.Peer | raw.base.InputPeer) -> int:
    """Get the non-raw peer id from a Peer object"""
    if isinstance(peer, (raw.types.PeerUser, raw.types.InputPeerUser)):
        return peer.user_id

    if isinstance(peer, (raw.types.PeerChat, raw.types.InputPeerChat)):
        return -peer.chat_id

    if isinstance(peer, (raw.types.PeerChannel, raw.types.InputPeerChannel)):
        return MAX_CHANNEL_ID - peer.channel_id

    if isinstance(peer, raw.types.InputPeerChannel):
        return MAX_CHANNEL_ID - peer.channel_id

    raise ValueError(f"Peer type invalid: {peer}")


def get_peer_type(peer_id: int) -> str:
    if peer_id < 0:
        if MIN_CHAT_ID <= peer_id:
            return "chat"

        if MIN_CHANNEL_ID <= peer_id < MAX_CHANNEL_ID:
            return "channel"
    elif 0 < peer_id <= MAX_USER_ID:
        return "user"

    raise ValueError(f"Peer id invalid: {peer_id}")


def get_channel_id(peer_id: int) -> int:
    return MAX_CHANNEL_ID - peer_id


def btoi(b: bytes) -> int:
    return int.from_bytes(b, "big")


def itob(i: int) -> bytes:
    return i.to_bytes(256, "big")


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def xor(a: bytes, b: bytes) -> bytes:
    return bytes(i ^ j for i, j in zip(a, b))


def compute_password_hash(
        algo: raw.types.PasswordKdfAlgoSHA256SHA256PBKDF2HMACSHA512iter100000SHA256ModPow,
        password: str
) -> bytes:
    hash1 = sha256(algo.salt1 + password.encode() + algo.salt1)
    hash2 = sha256(algo.salt2 + hash1 + algo.salt2)
    hash3 = hashlib.pbkdf2_hmac("sha512", hash2, algo.salt1, 100000)

    return sha256(algo.salt2 + hash3 + algo.salt2)


# noinspection PyPep8Naming
def compute_password_check(
        r: raw.types.account.Password,
        password: str
) -> raw.types.InputCheckPasswordSRP:
    algo = r.current_algo

    p_bytes = algo.p
    p = btoi(algo.p)

    g_bytes = itob(algo.g)
    g = algo.g

    B_bytes = r.srp_B
    B = btoi(B_bytes)

    srp_id = r.srp_id

    x_bytes = compute_password_hash(algo, password)
    x = btoi(x_bytes)

    g_x = pow(g, x, p)

    k_bytes = sha256(p_bytes + g_bytes)
    k = btoi(k_bytes)

    kg_x = (k * g_x) % p

    while True:
        a_bytes = os.urandom(256)
        a = btoi(a_bytes)

        A = pow(g, a, p)
        A_bytes = itob(A)

        u = btoi(sha256(A_bytes + B_bytes))

        if u > 0:
            break

    g_b = (B - kg_x) % p

    ux = u * x
    a_ux = a + ux
    S = pow(g_b, a_ux, p)
    S_bytes = itob(S)

    K_bytes = sha256(S_bytes)

    M1_bytes = sha256(
        xor(sha256(p_bytes), sha256(g_bytes))
        + sha256(algo.salt1)
        + sha256(algo.salt2)
        + A_bytes
        + B_bytes
        + K_bytes
    )

    return raw.types.InputCheckPasswordSRP(srp_id=srp_id, A=A_bytes, M1=M1_bytes)


async def parse_text_entities(
        client: "pylogram.Client",
        text: str,
        parse_mode: enums.ParseMode,
        entities: List["types.MessageEntity"]
) -> Dict[str, Union[str, List[raw.base.MessageEntity]]]:
    if entities:
        # Inject the client instance because parsing user mentions requires it
        for entity in entities:
            entity._client = client

        text, entities = text, [await entity.write() for entity in entities] or None
    else:
        text, entities = (await client.parser.parse(text, parse_mode)).values()

    return {
        "message": text,
        "entities": entities
    }


def zero_datetime() -> datetime:
    return datetime.fromtimestamp(0, timezone.utc)


def timestamp_to_datetime(ts: Optional[int]) -> Optional[datetime]:
    return datetime.fromtimestamp(ts) if ts else None


def datetime_to_timestamp(dt: Optional[datetime]) -> Optional[int]:
    return int(dt.timestamp()) if dt else None


@contextlib.asynccontextmanager
async def async_perf_counter(context: str = ""):
    start = time.perf_counter()

    try:
        yield
    finally:
        end = time.perf_counter()
        elapsed_time = end - start
        logging.info(f"[{context}] Elapsed time: {elapsed_time}")


def parse_username(username_string: str) -> str:
    username_string = username_string.lower()
    # Remove all valid user url components
    # like (https://username.t.me, https://t.me/username, https://telegram.me/username)
    username_string = re.sub(r"(http(s)?://)|(\.?t\.me/?)", "", username_string)
    # Remove all non-username characters
    username_string = re.sub(r"[^a-z0-9_]", "", username_string)

    return username_string


def parse_phone_number(phone_number: int | str) -> str | None:
    phone_number = re.sub(r"\D", "", str(phone_number))

    # Validate phone number format according to ISO 3166-1 alpha-2
    if not re.match(r"^\d{5,15}$", phone_number):
        return None

    return phone_number


def calculate_pagination_hash(numbers: List[int]) -> int:
    acc = 0

    for number in numbers:
        acc ^= acc >> 21
        acc ^= acc << 35
        acc ^= acc >> 4
        acc = (acc + number) & 0xffffffffffffffff  # To let it not overflow a 64-bit integer

    return acc


def is_tl_object_of_base_type(value: pylogram.raw.core.TLObject, base_type: Any | tuple[Any]) -> bool:
    if isinstance(base_type, tuple):
        return any(is_tl_object_of_base_type(value, t) for t in base_type)

    if base_type is None or base_type is Any:
        return True

    try:
        if isinstance(value, base_type):
            return True
    except TypeError:
        pass

    try:
        if issubclass(type(value), base_type):
            return True
    except TypeError:
        pass

    if base_type is type(value):
        return True

    origin = typing.get_origin(base_type)

    if origin is Union:
        for arg in typing.get_args(base_type):
            if is_tl_object_of_base_type(value, arg):
                return True

        return False

    return False


CHAT_LIST_INVITE_LINK_PREFIX: str = "https://t.me/addlist/"


def slug_to_chat_list_invite_link(slug: str) -> str:
    return f"{CHAT_LIST_INVITE_LINK_PREFIX}{slug}"


def chat_list_invite_link_to_slug(link: str) -> str:
    return link.replace(CHAT_LIST_INVITE_LINK_PREFIX, '')


def telegram_json_to_python_obj(data: raw.base.JSONValue) -> Any:
    if isinstance(data, raw.types.JsonObject):
        return {item.key: telegram_json_to_python_obj(item.value) for item in data.value}

    if isinstance(data, list):
        return [telegram_json_to_python_obj(item) for item in data]

    if isinstance(data, raw.types.JsonArray):
        return [telegram_json_to_python_obj(item) for item in data.value]

    if isinstance(data, raw.types.JsonNull):
        return None

    if isinstance(data, (raw.types.JsonBool, raw.types.JsonNumber, raw.types.JsonString)):
        return data.value

    raise ValueError(f"Unknown JSON object type: {data}")
